/*
 * dbcopy.s	- double word bcopy.
 *
 *	Fast memory mover - uses fp regs to do 8 byte at a time
 *	loads and stores !
 *
 *  Lots of ideas taken from Kipp Hickman's original version.
 *
 *  Written by: Gianni Mariani 19-Nov-1993
 */

	.weakext bcopy, _bcopy

#include <asm.h>
#include <regdef.h>

#define dst	a0	/* $4 */
#define src	a1	/* $5 */
#define size	a2	/* $6 */

#define LDW	l.d
#define	SDW	s.d

LEAF( _bcopy )

	move	v0, a1		    # move args around to be same as memcpy
	move	a1, a0
	move	a0, v0
	j	memcpy_start
	
XNESTED( memmove )
	move    v0, a0	
	j	memcpy_start

XNESTED( memcpy )
	move	v0, a0

memcpy_start:

	subu    sp, 80		    # make some scratch space
	beq	src, dst, ret	    # get out when src == dst
	
/*  if ( ( dst > src ) && ( dst < ( src + size ) ) ) goto backwards;	*/
	bltu	dst, src, forwards
	addu	v1, src, size
	bltu	dst, v1, backwards

forwards:
/*  if ( size < 4 * sizeof( d4_t ) ) goto finishit;			*/
	bltu	size, 16, finishit
	
/*  v4 = ( ( d4_t ) dst ) & ( sizeof( d8_t ) - 1 );			*/
/*  v5 = ( ( d4_t ) src ) & ( sizeof( d8_t ) - 1 );			*/
/*  v7 = ( ( d4_t ) src ) & ( sizeof( d4_t ) - 1 );			*/
/*  if ( v4 == v5 ) goto aligned_forw_8_copy;				*/

	and	t5, src, 7
	and	t4, dst, 7
	and	t7, src, 3
	beq	t5, t4, aligned_forw_8_copy
	
/*  v6 = ( ( d4_t ) dst ) & ( sizeof( d4_t ) - 1 );			*/
/*  if ( v6 != v7 ) goto unaligned_forward;				*/
	and	t6, dst, 3
	bne	t6, t7, unaligned_forward

/* ==================================================================== */
/* ==================================================================== */
/* Copy forward - src and dst are only 4 byte aligned			*/

aligned_forw_4_copy :

/*  Make src aligned on word -						*/
/*  if ( v7 ) {								*/
/*  ( ( ud4_t * ) dst )[ 0 ] = ( ( ud4_t * ) src )[ 0 ];		*/

	and	t5, t5, 4
	beq	t7, 0, 33f	# align src only if nessasary
	lwl	v1, 0(src)	# read only enough to align
	swl	v1, 0(dst)	# write alignment
	
/*  v7 = sizeof( d4_t ) - v7;						*/
/*  dst += v7;								*/
/*  src += v7;								*/
/*  size -= v7;								*/
/*  v5 = ( ( d4_t ) src ) & ( sizeof( d8_t ) - 1 );			*/

	addu	size, size, t7
	subu	dst, dst, t7
	subu	src, src, t7
	
	addu	src, src, 4
	and	t5, src, 4
	subu	size, size, 4
	addu	dst, dst, 4
33:

/*  Double word align the src, dst is NOT double aligned		*/
/*  if ( v5 ) {								*/
/*  ( ( d4_t * ) dst )[ 0 ] = ( ( d4_t * ) src )[ 0 ];			*/

	lw	v1, 0(src)
	beq	t5, 0, 34f
	sw	v1, 0(dst)
	
	subu	size, size, 4
	addu	dst, dst, 4
	addu	src, src, 4
34:


/*  if ( size > 4 *sizeof( ud128_t ) ) {				*/

	bltu	size, 512, forw_4_64_loop
	

	/* Why - you might ask - do we bother going all the way to 128	*/
	/* bytes - This is the magic secondary cache size.		*/
	/* cache thrashes will be limited to at most two per cache line	*/

	/* save fp regs that need saving				*/
	SDW	$f30, 40(sp)
	SDW	$f28, 32(sp)
	SDW	$f26, 24(sp)
	SDW	$f24, 16(sp)
	SDW	$f22, 8(sp)
	SDW	$f20, 0(sp)

	/* while ( size > sizeof( ud128_t ) ) {				*/
	/* do the copy - 128 bytes at a time				*/
57:
	LDW	$f0 ,  0(src)
	LDW	$f2 ,  8(src)
	LDW	$f4 , 16(src)
	LDW	$f6 , 24(src)
	LDW	$f8 , 32(src)
	LDW	$f10, 40(src)
	LDW	$f12, 48(src)
	LDW	$f14, 56(src)
	LDW	$f16, 64(src)
	LDW	$f18, 72(src)
	LDW	$f20, 80(src)
	LDW	$f22, 88(src)
	LDW	$f24, 96(src)
	LDW	$f26,104(src)
	LDW	$f28,112(src)
	LDW	$f30,120(src)

	swc1	$f1 ,  0(dst)
	swc1	$f0 ,  4(dst)
	swc1	$f3 ,  8(dst)
	swc1	$f2 , 12(dst)
	swc1	$f5 , 16(dst)
	swc1	$f4 , 20(dst)
	swc1	$f7 , 24(dst)
	swc1	$f6 , 28(dst)
	swc1	$f9 , 32(dst)
	swc1	$f8 , 36(dst)
	swc1	$f11, 40(dst)
	swc1	$f10 ,44(dst)
	swc1	$f13 ,48(dst)
	swc1	$f12 ,52(dst)
	swc1	$f15 ,56(dst)
	swc1	$f14 ,60(dst)
	swc1	$f17 ,64(dst)
	swc1	$f16 ,68(dst)
	swc1	$f19 ,72(dst)
	swc1	$f18 ,76(dst)
	swc1	$f21 ,80(dst)
	swc1	$f20 ,84(dst)
	swc1	$f23 ,88(dst)
	swc1	$f22 ,92(dst)
	swc1	$f25 ,96(dst)
	swc1	$f24 ,100(dst)
	swc1	$f27 ,104(dst)
	swc1	$f26 ,108(dst)
	swc1	$f29 ,112(dst)
	swc1	$f28 ,116(dst)
	swc1	$f31 ,120(dst)
	swc1	$f30 ,124(dst)
	
	subu	size, size, 128
	addu	dst, dst, 128
	addu	src, src, 128
	bgeu	size, 128, 57b

	LDW	$f30, 40(sp)
	LDW	$f28, 32(sp)
	LDW	$f26, 24(sp)
	LDW	$f24, 16(sp)
	LDW	$f22, 8(sp)
	LDW	$f20, 0(sp)

/* while ( size > sizeof( ud64_t ) ) {					*/
	
forw_4_64_loop:
	bltu	size, 64, forw_4_32_if
57:
	LDW	$f0 ,  0(src)
	LDW	$f2 ,  8(src)
	LDW	$f4 , 16(src)
	LDW	$f6 , 24(src)
	LDW	$f8 , 32(src)
	LDW	$f10, 40(src)
	LDW	$f12, 48(src)
	LDW	$f14, 56(src)

	swc1	$f1 ,  0(dst)
	swc1	$f0 ,  4(dst)
	swc1	$f3 ,  8(dst)
	swc1	$f2 , 12(dst)
	swc1	$f5 , 16(dst)
	swc1	$f4 , 20(dst)
	swc1	$f7 , 24(dst)
	swc1	$f6 , 28(dst)
	swc1	$f9 , 32(dst)
	swc1	$f8 , 36(dst)
	swc1	$f11, 40(dst)
	swc1	$f10 ,44(dst)
	swc1	$f13 ,48(dst)
	swc1	$f12 ,52(dst)
	swc1	$f15 ,56(dst)
	swc1	$f14 ,60(dst)
	
	subu	size, size, 64
	addu	dst, dst, 64
	addu	src, src, 64
	bgeu	size, 64, 57b
	
forw_4_32_if:

/*  if ( size > sizeof( ud32_t ) ) {					*/
	bltu	size, 32, forw_4_16_if
	
	LDW	$f0 ,  0(src)
	LDW	$f2 ,  8(src)
	LDW	$f4 , 16(src)
	LDW	$f6 , 24(src)

	swc1	$f1 ,  0(dst)
	swc1	$f0 ,  4(dst)
	swc1	$f3 ,  8(dst)
	swc1	$f2 , 12(dst)
	swc1	$f5 , 16(dst)
	swc1	$f4 , 20(dst)
	swc1	$f7 , 24(dst)
	swc1	$f6 , 28(dst)
	
	subu	size, size, 32
	addu	dst, dst, 32
	addu	src, src, 32
	
forw_4_16_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 16, forw_4_8_if
	
	LDW	$f0 ,  0(src)
	LDW	$f2 ,  8(src)

	swc1	$f1 ,  0(dst)
	swc1	$f0 ,  4(dst)
	swc1	$f3 ,  8(dst)
	swc1	$f2 , 12(dst)
	
	subu	size, size, 16
	addu	dst, dst, 16
	addu	src, src, 16

forw_4_8_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 8, aligned_forw_finish
	
	LDW	$f0 ,  0(src)

	swc1	$f1 ,  0(dst)
	swc1	$f0 ,  4(dst)
	
	subu	size, size, 8
	addu	dst, dst, 8
	addu	src, src, 8

aligned_forw_finish:

/*  aligned_forw_finish :						*/
/*  while ( size >= sizeof( d4_t ) ) {					*/
	bltu	size, 4, aligned_forw_partial
63:

	lw	t0, 0(src)
	sw	t0, 0(dst)

	subu	size, size, 4
	addu	src, src, 4
	addu	dst, dst, 4
	bgeu	size, 4, 63b

aligned_forw_partial:

	beq     size, 0, ret

	addu	src, src, size
	addu	dst, dst, size

	lwr	t0, -1(src)
	swr	t0, -1(dst)

ret:
	addu	sp, 80
	j	ra

/* ==================================================================== */
/* ==================================================================== */
/* Copy forward - src and dst are 8 byte aligned			*/

    
aligned_forw_8_copy:

/*  Make src aligned on word -						*/
/*  if ( v7 ) {								*/
/*  ( ( ud4_t * ) dst )[ 0 ] = ( ( ud4_t * ) src )[ 0 ];		*/

	and	t5, t5, 4
	beq	t7, 0, 33f	# align src only if nessasary
	lwl	v1, 0(src)	# read only enough to align
	swl	v1, 0(dst)	# write alignment
	
/*  v7 = sizeof( d4_t ) - v7;						*/
/*  dst += v7;								*/
/*  src += v7;								*/
/*  size -= v7;								*/
/*  v5 = ( ( d4_t ) src ) & ( sizeof( d8_t ) - 1 );			*/

	addu	size, size, t7
	subu	dst, dst, t7
	subu	src, src, t7
	
	addu	src, src, 4
	and	t5, src, 4
	subu	size, size, 4
	addu	dst, dst, 4
33:

/*  Double word align the src, dst is NOT double aligned		*/
/*  if ( v5 ) {								*/
/*  ( ( d4_t * ) dst )[ 0 ] = ( ( d4_t * ) src )[ 0 ];			*/

	lw	v1, 0(src)
	beq	t5, 0, 34f
	sw	v1, 0(dst)
	
	subu	size, size, 4
	addu	dst, dst, 4
	addu	src, src, 4
34:


/*  if ( size > 4 *sizeof( ud128_t ) ) {				*/

	bltu	size, 512, forw_8_64_loop

	/* Why - you might ask - do we bother going all the way to 128	*/
	/* bytes - This is the magic secondary cache line size.		*/
	/* cache thrashes will be limited to at most two per cache line	*/

	/* save fp regs that need saving				*/
	SDW	$f30, 40(sp)
	SDW	$f28, 32(sp)
	SDW	$f26, 24(sp)
	SDW	$f24, 16(sp)
	SDW	$f22, 8(sp)
	SDW	$f20, 0(sp)

	/* while ( size > sizeof( ud128_t ) ) {				*/
	/* do the copy - 128 bytes at a time				*/
57:
	LDW	$f0 ,  0(src)
	LDW	$f2 ,  8(src)
	LDW	$f4 , 16(src)
	LDW	$f6 , 24(src)
	LDW	$f8 , 32(src)
	LDW	$f10, 40(src)
	LDW	$f12, 48(src)
	LDW	$f14, 56(src)
	LDW	$f16, 64(src)
	LDW	$f18, 72(src)
	LDW	$f20, 80(src)
	LDW	$f22, 88(src)
	LDW	$f24, 96(src)
	LDW	$f26,104(src)
	LDW	$f28,112(src)
	LDW	$f30,120(src)

	SDW	$f0 ,  0(dst)
	SDW	$f2 ,  8(dst)
	SDW	$f4 , 16(dst)
	SDW	$f6 , 24(dst)
	SDW	$f8 , 32(dst)
	SDW	$f10, 40(dst)
	SDW	$f12, 48(dst)
	SDW	$f14, 56(dst)
	SDW	$f16, 64(dst)
	SDW	$f18, 72(dst)
	SDW	$f20, 80(dst)
	SDW	$f22, 88(dst)
	SDW	$f24, 96(dst)
	SDW	$f26,104(dst)
	SDW	$f28,112(dst)
	SDW	$f30,120(dst)

	subu	size, size, 128
	addu	dst, dst, 128
	addu	src, src, 128
	bgeu	size, 128, 57b

	LDW	$f30, 40(sp)
	LDW	$f28, 32(sp)
	LDW	$f26, 24(sp)
	LDW	$f24, 16(sp)
	LDW	$f22, 8(sp)
	LDW	$f20, 0(sp)

/* while ( size > sizeof( ud64_t ) ) {					*/
	
forw_8_64_loop:
	bltu	size, 64, forw_8_32_if
57:
	LDW	$f0 ,  0(src)
	LDW	$f2 ,  8(src)
	LDW	$f4 , 16(src)
	LDW	$f6 , 24(src)
	LDW	$f8 , 32(src)
	LDW	$f10, 40(src)
	LDW	$f12, 48(src)
	LDW	$f14, 56(src)

	SDW	$f0 ,  0(dst)
	SDW	$f2 ,  8(dst)
	SDW	$f4 , 16(dst)
	SDW	$f6 , 24(dst)
	SDW	$f8 , 32(dst)
	SDW	$f10, 40(dst)
	SDW	$f12, 48(dst)
	SDW	$f14, 56(dst)
	
	subu	size, size, 64
	addu	dst, dst, 64
	addu	src, src, 64
	bgeu	size, 64, 57b
	
forw_8_32_if:

/*  if ( size > sizeof( ud32_t ) ) {					*/
	bltu	size, 32, forw_8_16_if
	
	LDW	$f0 ,  0(src)
	LDW	$f2 ,  8(src)
	LDW	$f4 , 16(src)
	LDW	$f6 , 24(src)

	SDW	$f0 ,  0(dst)
	SDW	$f2 ,  8(dst)
	SDW	$f4 , 16(dst)
	SDW	$f6 , 24(dst)

	subu	size, size, 32
	addu	dst, dst, 32
	addu	src, src, 32
	
forw_8_16_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 16, forw_8_8_if
	
	LDW	$f0 ,  0(src)
	LDW	$f2 ,  8(src)

	SDW	$f0 ,  0(dst)
	SDW	$f2 ,  8(dst)

	subu	size, size, 16
	addu	dst, dst, 16
	addu	src, src, 16

forw_8_8_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 8, aligned_forw_finish
	
	LDW	$f0 ,  0(src)
	SDW	$f0 ,  0(dst)

	subu	size, size, 8
	addu	dst, dst, 8
	addu	src, src, 8

/*  goto aligned_forw_finish;						*/

	b	aligned_forw_finish

/* ==================================================================== */
/* ==================================================================== */
/* Copy forward - src and dst are not aligned				*/

unaligned_forward:

	beq	t7, 0, 33f	# align src only if nessasary
	beq     t7, 1, 23f
	beq     t7, 2, 22f

	lbu	t0, 0(src)	# copy one byte
	sb	t0, 0(dst)
	
	subu	size, size, 1
	addu	dst, dst, 1
	addu	src, src, 1

	b	33f

23:
	lbu	t0, 0(src)	# copy three bytes
	lbu	t1, 1(src)
	lbu	t2, 2(src)
	sb	t0, 0(dst)
	sb	t1, 1(dst)
	sb	t2, 2(dst)
	subu	size, size, 3
	addu	dst, dst, 3
	addu	src, src, 3

	b	33f

22:

	lbu	t0, 0(src)	# copy two bytes
	lbu	t1, 1(src)
	sb	t0, 0(dst)
	sb	t1, 1(dst)
	subu	size, size, 2
	addu	dst, dst, 2
	addu	src, src, 2


/*  if ( size > 4 *sizeof( ud64_t ) ) {				*/
33:
	bltu	size, 256, forw_U_32_loop

	/* save regs that need saving					*/
	sw	s0,   0(sp)
	sw	s1,   4(sp)
	sw	s2,   8(sp)
	sw	s3,  12(sp)

	/* while ( size >= sizeof( ud64_t ) ) {				*/
	/* do the copy - 64 bytes at a time				*/
57:
	lw	t0 ,  0(src)
	lw	t1 ,  4(src)
	lw	t2 ,  8(src)
	lw	t3 , 12(src)
	lw	t4 , 16(src)
	lw	t5 , 20(src)
	lw	t6 , 24(src)
	lw	t7 , 28(src)
	lw	s0 , 32(src)
	lw	s1 , 36(src)
	lw	s2 , 40(src)
	lw	s3 , 44(src)
	lw	v1 , 48(src)
	lw	a3 , 52(src)
	lw	t8 , 56(src)
	lw	t9 , 60(src)
	
	usw	t0 ,  0(dst)
	usw	t1 ,  4(dst)
	usw	t2 ,  8(dst)
	usw	t3 , 12(dst)
	usw	t4 , 16(dst)
	usw	t5 , 20(dst)
	usw	t6 , 24(dst)
	usw	t7 , 28(dst)
	usw	s0 , 32(dst)
	usw	s1 , 36(dst)
	usw	s2 , 40(dst)
	usw	s3 , 44(dst)
	usw	v1 , 48(dst)
	usw	a3 , 52(dst)
	usw	t8 , 56(dst)
	usw	t9 , 60(dst)

	subu	size, size, 64
	addu	dst, dst, 64
	addu	src, src, 64
	bgeu	size, 64, 57b

	lw	s0,   0(sp)
	lw	s1,   4(sp)
	lw	s2,   8(sp)
	lw	s3,  12(sp)

/* while ( size > sizeof( ud64_t ) ) {					*/
	
forw_U_32_loop:
	bltu	size, 32, forw_U_16_if

5:
	lw	t0 ,  0(src)
	lw	t1 ,  4(src)
	lw	t2 ,  8(src)
	lw	t3 , 12(src)
	lw	t4 , 16(src)
	lw	t5 , 20(src)
	lw	t6 , 24(src)
	lw	t7 , 28(src)
	
	usw	t0 ,  0(dst)
	usw	t1 ,  4(dst)
	usw	t2 ,  8(dst)
	usw	t3 , 12(dst)
	usw	t4 , 16(dst)
	usw	t5 , 20(dst)
	usw	t6 , 24(dst)
	usw	t7 , 28(dst)

	subu	size, size, 32
	addu	dst, dst, 32
	addu	src, src, 32
	bgeu	size, 32, 5b

forw_U_16_if:
	bltu	size, 16, forw_U_8_if

	lw	t0 ,  0(src)
	lw	t1 ,  4(src)
	lw	t2 ,  8(src)
	lw	t3 , 12(src)
	
	usw	t0 ,  0(dst)
	usw	t1 ,  4(dst)
	usw	t2 ,  8(dst)
	usw	t3 , 12(dst)

	subu	size, size, 16
	addu	dst, dst, 16
	addu	src, src, 16

forw_U_8_if:
	bltu	size, 8, finishit

	lw	t0 ,  0(src)
	lw	t1 ,  4(src)
	
	usw	t0 ,  0(dst)
	usw	t1 ,  4(dst)

	subu	size, size, 8
	addu	dst, dst, 8
	addu	src, src, 8

finishit:

/*  while ( size > sizeof( d4_t ) ) {					*/
/*  ( ( ud4_t * ) dst )[ 0 ] = ( ( ud4_t * ) src )[ 0 ];		*/

	bltu	size, 4, finishit_byte
4:
	ulw	t0, 0(src)
	usw	t0, 0(dst)

	subu	size, size, 4
	addu	dst, dst, 4
	addu	src, src, 4
	
	bgeu	size, 4, 4b

finishit_byte:

/*  while ( size > 0 ) {						*/
	beq	size, 0, ret
	beq	size, 1, 71f
	beq	size, 2, 72f

	lbu	t0,0(src)	# copy three bytes
	lbu	t1,1(src)
	lbu	t2,2(src)
	sb	t0,0(dst)
	sb	t1,1(dst)
	sb	t2,2(dst)

	b	ret
72:
	
	lbu	t0,0(src)	# copy two bytes
	lbu	t1,1(src)
	sb	t0,0(dst)
	sb	t1,1(dst)

	b	ret

71:
	lbu	t0, 0(src)
	sb	t0, 0(dst)
	
	b	ret

/* ==================================================================== */
/* ==================================================================== */
/* Backwards copies							*/

backwards:

/* Point to the end							*/
	addu	dst, dst, size
	addu	src, src, size

/*  if ( size < 4 * sizeof( d4_t ) ) goto finishit;			*/
	bltu	size, 16, finishit_backw
	
/*  v4 = ( ( d4_t ) dst ) & ( sizeof( d8_t ) - 1 );			*/
/*  v5 = ( ( d4_t ) src ) & ( sizeof( d8_t ) - 1 );			*/
/*  v7 = ( ( d4_t ) src ) & ( sizeof( d4_t ) - 1 );			*/
/*  if ( v4 == v5 ) goto aligned_backw_8_copy;				*/

	and	t5, src, 7
	and	t4, dst, 7
	and	t7, src, 3
	beq	t5, t4, aligned_backw_8_copy
	
/*  v6 = ( ( d4_t ) dst ) & ( sizeof( d4_t ) - 1 );			*/
/*  if ( v6 != v7 ) goto unaligned_backward;				*/
	and	t6, dst, 3
	bne	t6, t7, unaligned_backward

/* ==================================================================== */
/* ==================================================================== */
/* Copy backward - src and dst are only 4 byte aligned			*/

aligned_back_4_copy :

/*  Make src aligned on word -						*/
/*  if ( v7 ) {								*/
/*  ( ( ud4_t * ) dst )[ 0 ] = ( ( ud4_t * ) src )[ 0 ];		*/

	and	t5, t5, 4
	beq	t7, 0, 33f	# align src only if nessasary
	lwr	v1, -1(src)	# read only enough to align
	swr	v1, -1(dst)	# write alignment

/*  v7 = sizeof( d4_t ) - v7;						*/
/*  dst += v7;								*/
/*  src += v7;								*/
/*  size -= v7;								*/
/*  v5 = ( ( d4_t ) src ) & ( sizeof( d8_t ) - 1 );			*/

	subu	size, size, t7
	subu	dst, dst, t7
	subu	src, src, t7
	
	and	t5, src, 4
33:

/*  Double word align the src, dst is NOT double aligned		*/
/*  if ( v5 ) {								*/
/*  ( ( d4_t * ) dst )[ 0 ] = ( ( d4_t * ) src )[ 0 ];			*/

	lw	v1, -4(src)
	beq	t5, 0, 34f
	sw	v1, -4(dst)
	
	subu	size, size, 4
	subu	dst, dst, 4
	subu	src, src, 4
34:


/*  if ( size > 4 *sizeof( ud128_t ) ) {				*/

	bltu	size, 512, backw_4_64_loop
	

	/* Why - you might ask - do we bother going all the way to 128	*/
	/* bytes - This is the magic secondary cache size.		*/
	/* cache thrashes will be limited to at most two per cache line	*/

	/* save fp regs that need saving				*/
	SDW	$f30, 40(sp)
	SDW	$f28, 32(sp)
	SDW	$f26, 24(sp)
	SDW	$f24, 16(sp)
	SDW	$f22, 8(sp)
	SDW	$f20, 0(sp)

	/* while ( size > sizeof( ud128_t ) ) {				*/
	/* do the copy - 128 bytes at a time				*/
57:
	LDW	$f0 , -  8(src)
	LDW	$f2 , - 16(src)
	LDW	$f4 , - 24(src)
	LDW	$f6 , - 32(src)
	LDW	$f8 , - 40(src)
	LDW	$f10, - 48(src)
	LDW	$f12, - 56(src)
	LDW	$f14, - 64(src)
	LDW	$f16, - 72(src)
	LDW	$f18, - 80(src)
	LDW	$f20, - 88(src)
	LDW	$f22, - 96(src)
	LDW	$f24, -104(src)
	LDW	$f26, -112(src)
	LDW	$f28, -120(src)
	LDW	$f30, -128(src)

	swc1	$f0 , -  4(dst)
	swc1	$f1 , -  8(dst)
	swc1	$f2 , - 12(dst)
	swc1	$f3 , - 16(dst)
	swc1	$f4 , - 20(dst)
	swc1	$f5 , - 24(dst)
	swc1	$f6 , - 28(dst)
	swc1	$f7 , - 32(dst)
	swc1	$f8 , - 36(dst)
	swc1	$f9 , - 40(dst)
	swc1	$f10 , -44(dst)
	swc1	$f11 , -48(dst)
	swc1	$f12 , -52(dst)
	swc1	$f13 , -56(dst)
	swc1	$f14 , -60(dst)
	swc1	$f15 , -64(dst)
	swc1	$f16 , -68(dst)
	swc1	$f17 , -72(dst)
	swc1	$f18 , -76(dst)
	swc1	$f19 , -80(dst)
	swc1	$f20 , -84(dst)
	swc1	$f21 , -88(dst)
	swc1	$f22 , -92(dst)
	swc1	$f23 , -96(dst)
	swc1	$f24 , -100(dst)
	swc1	$f25 , -104(dst)
	swc1	$f26 , -108(dst)
	swc1	$f27 , -112(dst)
	swc1	$f28 , -116(dst)
	swc1	$f29 , -120(dst)
	swc1	$f30 , -124(dst)
	swc1	$f31 , -128(dst)
	
	subu	size, size, 128
	subu	dst, dst, 128
	subu	src, src, 128
	bgeu	size, 128, 57b

	LDW	$f30, 40(sp)
	LDW	$f28, 32(sp)
	LDW	$f26, 24(sp)
	LDW	$f24, 16(sp)
	LDW	$f22, 8(sp)
	LDW	$f20, 0(sp)

/* while ( size > sizeof( ud64_t ) ) {					*/
	
backw_4_64_loop:
	bltu	size, 64, backw_4_32_if
57:
	LDW	$f0 , -  8(src)
	LDW	$f2 , - 16(src)
	LDW	$f4 , - 24(src)
	LDW	$f6 , - 32(src)
	LDW	$f8 , - 40(src)
	LDW	$f10, - 48(src)
	LDW	$f12, - 56(src)
	LDW	$f14, - 64(src)

	swc1	$f0 , -  4(dst)
	swc1	$f1 , -  8(dst)
	swc1	$f2 , - 12(dst)
	swc1	$f3 , - 16(dst)
	swc1	$f4 , - 20(dst)
	swc1	$f5 , - 24(dst)
	swc1	$f6 , - 28(dst)
	swc1	$f7 , - 32(dst)
	swc1	$f8 , - 36(dst)
	swc1	$f9 , - 40(dst)
	swc1	$f10 , -44(dst)
	swc1	$f11 , - 48(dst)
	swc1	$f12 , - 52(dst)
	swc1	$f13 , - 56(dst)
	swc1	$f14 , - 60(dst)
	swc1	$f15 , - 64(dst)
	
	subu	size, size, 64
	subu	dst, dst, 64
	subu	src, src, 64
	bgeu	size, 64, 57b
	
backw_4_32_if:

/*  if ( size > sizeof( ud32_t ) ) {					*/
	bltu	size, 32, backw_4_16_if
	
	LDW	$f0 , -  8(src)
	LDW	$f2 , - 16(src)
	LDW	$f4 , - 24(src)
	LDW	$f6 , - 32(src)

	swc1	$f0 , -  4(dst)
	swc1	$f1 , -  8(dst)
	swc1	$f2 , - 12(dst)
	swc1	$f3 , - 16(dst)
	swc1	$f4 , - 20(dst)
	swc1	$f5 , - 24(dst)
	swc1	$f6 , - 28(dst)
	swc1	$f7 , - 32(dst)
	
	subu	size, size, 32
	subu	dst, dst, 32
	subu	src, src, 32
	
backw_4_16_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 16, backw_4_8_if
	
	LDW	$f0 , -  8(src)
	LDW	$f2 , - 16(src)

	swc1	$f0 , -  4(dst)
	swc1	$f1 , -  8(dst)
	swc1	$f2 , - 12(dst)
	swc1	$f3 , - 16(dst)
	
	subu	size, size, 16
	subu	dst, dst, 16
	subu	src, src, 16

backw_4_8_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 8, aligned_backw_finish
	
	LDW	$f0 , - 8(src)

	swc1	$f0 , - 4(dst)
	swc1	$f1 , - 8(dst)
	
	subu	size, size, 8
	subu	dst, dst, 8
	subu	src, src, 8

aligned_backw_finish:
/*  aligned_backw_finish :						*/
/*  while ( size >= sizeof( d4_t ) ) {					*/
	bltu	size, 4, aligned_backw_partial
63:

	lw	t0, - 4(src)
	sw	t0, - 4(dst)

	subu	size, size, 4
	subu	src, src, 4
	subu	dst, dst, 4
	bgeu	size, 4, 63b

aligned_backw_partial:

	beq     size, 0, ret

	subu	src, src, size
	subu	dst, dst, size

	lwl	t0, 0(src)
	swl	t0, 0(dst)

	b	ret

/* ==================================================================== */
/* ==================================================================== */
/* Copy backward - src and dst are 8 byte aligned			*/

    
aligned_backw_8_copy:

/*  Make src aligned on word -						*/
/*  if ( v7 ) {								*/
/*  ( ( ud4_t * ) dst )[ 0 ] = ( ( ud4_t * ) src )[ 0 ];		*/

	and	t5, t5, 4
	beq	t7, 0, 33f	# align src only if nessasary
	lwr	v1, -1(src)	# read only enough to align
	swr	v1, -1(dst)	# write alignment
	
/*  v7 = sizeof( d4_t ) - v7;						*/
/*  dst += v7;								*/
/*  src += v7;								*/
/*  size -= v7;								*/
/*  v5 = ( ( d4_t ) src ) & ( sizeof( d8_t ) - 1 );			*/

	subu	size, size, t7
	subu	dst, dst, t7
	subu	src, src, t7
	
	and	t5, src, 4
33:

/*  Double word align the src, dst is NOT double aligned		*/
/*  if ( v5 ) {								*/
/*  ( ( d4_t * ) dst )[ 0 ] = ( ( d4_t * ) src )[ 0 ];			*/

	lw	v1, -4(src)
	beq	t5, 0, 34f
	sw	v1, -4(dst)
	
	subu	size, size, 4
	subu	dst, dst, 4
	subu	src, src, 4

34:

/*  if ( size > 4 *sizeof( ud128_t ) ) {				*/

	bltu	size, 512, backw_8_64_loop

	/* Why - you might ask - do we bother going all the way to 128	*/
	/* bytes - This is the magic secondary cache line size.		*/
	/* cache thrashes will be limited to at most two per cache line	*/

	/* save fp regs that need saving				*/
	SDW	$f30, 40(sp)
	SDW	$f28, 32(sp)
	SDW	$f26, 24(sp)
	SDW	$f24, 16(sp)
	SDW	$f22, 8(sp)
	SDW	$f20, 0(sp)

	/* while ( size > sizeof( ud128_t ) ) {				*/
	/* do the copy - 128 bytes at a time				*/
57:
	LDW	$f0 , -  8(src)
	LDW	$f2 , - 16(src)
	LDW	$f4 , - 24(src)
	LDW	$f6 , - 32(src)
	LDW	$f8 , - 40(src)
	LDW	$f10, - 48(src)
	LDW	$f12, - 56(src)
	LDW	$f14, - 64(src)
	LDW	$f16, - 72(src)
	LDW	$f18, - 80(src)
	LDW	$f20, - 88(src)
	LDW	$f22, - 96(src)
	LDW	$f24, -104(src)
	LDW	$f26, -112(src)
	LDW	$f28, -120(src)
	LDW	$f30, -128(src)

	SDW	$f0 ,  - 8(dst)
	SDW	$f2 ,  -16(dst)
	SDW	$f4 ,  -24(dst)
	SDW	$f6 ,  -32(dst)
	SDW	$f8 ,  -40(dst)
	SDW	$f10,  -48(dst)
	SDW	$f12, - 56(dst)
	SDW	$f14, - 64(dst)
	SDW	$f16, - 72(dst)
	SDW	$f18, - 80(dst)
	SDW	$f20, - 88(dst)
	SDW	$f22, - 96(dst)
	SDW	$f24, -104(dst)
	SDW	$f26, -112(dst)
	SDW	$f28, -120(dst)
	SDW	$f30, -128(dst)

	subu	size, size, 128
	subu	dst, dst, 128
	subu	src, src, 128
	bgeu	size, 128, 57b

	LDW	$f30, 40(sp)
	LDW	$f28, 32(sp)
	LDW	$f26, 24(sp)
	LDW	$f24, 16(sp)
	LDW	$f22, 8(sp)
	LDW	$f20, 0(sp)

/* while ( size > sizeof( ud64_t ) ) {					*/
	
backw_8_64_loop:
	bltu	size, 64, backw_8_32_if
57:
	LDW	$f0 , -  8(src)
	LDW	$f2 , - 16(src)
	LDW	$f4 , - 24(src)
	LDW	$f6 , - 32(src)
	LDW	$f8 , - 40(src)
	LDW	$f10, - 48(src)
	LDW	$f12, - 56(src)
	LDW	$f14, - 64(src)

	SDW	$f0 , -  8(dst)
	SDW	$f2 , - 16(dst)
	SDW	$f4 , - 24(dst)
	SDW	$f6 , - 32(dst)
	SDW	$f8 , - 40(dst)
	SDW	$f10, - 48(dst)
	SDW	$f12, - 56(dst)
	SDW	$f14, - 64(dst)
	
	subu	size, size, 64
	subu	dst, dst, 64
	subu	src, src, 64
	bgeu	size, 64, 57b
	
backw_8_32_if:

/*  if ( size > sizeof( ud32_t ) ) {					*/
	bltu	size, 32, backw_8_16_if
	
	LDW	$f0 , -  8(src)
	LDW	$f2 , - 16(src)
	LDW	$f4 , - 24(src)
	LDW	$f6 , - 32(src)

	SDW	$f0 , -  8(dst)
	SDW	$f2 , - 16(dst)
	SDW	$f4 , - 24(dst)
	SDW	$f6 , - 32(dst)

	subu	size, size, 32
	subu	dst, dst, 32
	subu	src, src, 32
	
backw_8_16_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 16, backw_8_8_if
	
	LDW	$f0 , -  8(src)
	LDW	$f2 , - 16(src)

	SDW	$f0 , -  8(dst)
	SDW	$f2 , - 16(dst)

	subu	size, size, 16
	subu	dst, dst, 16
	subu	src, src, 16

backw_8_8_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 8, aligned_backw_finish
	
	LDW	$f0 , -8(src)
	SDW	$f0 , -8(dst)

	subu	size, size, 8
	subu	dst, dst, 8
	subu	src, src, 8

/*  goto aligned_backw_finish;						*/

	b	aligned_backw_finish

/* ==================================================================== */
/* ==================================================================== */
/* Copy backward - src and dst are not aligned				*/

unaligned_backward:

/*  Make src aligned on word -						*/
/*  if ( v7 ) {								*/
/*  ( ( ud4_t * ) dst )[ 0 ] = ( ( ud4_t * ) src )[ 0 ];		*/

	beq	t7, 0, 33f	# align src only if nessasary
	beq     t7, 3, 23f
	beq     t7, 2, 22f

	lbu	t0, -1(src)	# copy one byte
	sb	t0, -1(dst)
	
	subu	size, size, 1
	subu	dst, dst, 1
	subu	src, src, 1

	b	33f

23:
	lbu	t0,-1(src)	# copy three bytes
	lbu	t1,-2(src)
	lbu	t2,-3(src)
	sb	t0,-1(dst)
	sb	t1,-2(dst)
	sb	t2,-3(dst)
	subu	size, size, 3
	subu	dst, dst, 3
	subu	src, src, 3

	b	33f

22:

	lbu	t0,-1(src)	# copy three bytes
	lbu	t1,-2(src)
	sb	t0,-1(dst)
	sb	t1,-2(dst)
	subu	size, size, 2
	subu	dst, dst, 2
	subu	src, src, 2



/*  if ( size > 4 *sizeof( ud64_t ) ) {				*/
33:
	bltu	size, 256, backw_U_32_loop

	/* save regs that need saving					*/
	sw	s0,   0(sp)
	sw	s1,   4(sp)
	sw	s2,   8(sp)
	sw	s3,  12(sp)

	/* while ( size >= sizeof( ud64_t ) ) {				*/
	/* do the copy - 64 bytes at a time				*/
57:
	lw	t0 , -  4(src)
	lw	t1 , -  8(src)
	lw	t2 , - 12(src)
	lw	t3 , - 16(src)
	lw	t4 , - 20(src)
	lw	t5 , - 24(src)
	lw	t6 , - 28(src)
	lw	t7 , - 32(src)
	lw	s0 , - 36(src)
	lw	s1 , - 40(src)
	lw	s2 , - 44(src)
	lw	s3 , - 48(src)
	lw	v1 , - 52(src)
	lw	a3 , - 56(src)
	lw	t8 , - 60(src)
	lw	t9 , - 64(src)
	
	usw	t0 , -  4(dst)
	usw	t1 , -  8(dst)
	usw	t2 , - 12(dst)
	usw	t3 , - 16(dst)
	usw	t4 , - 20(dst)
	usw	t5 , - 24(dst)
	usw	t6 , - 28(dst)
	usw	t7 , - 32(dst)
	usw	s0 , - 36(dst)
	usw	s1 , - 40(dst)
	usw	s2 , - 44(dst)
	usw	s3 , - 48(dst)
	usw	v1 , - 52(dst)
	usw	a3 , - 56(dst)
	usw	t8 , - 60(dst)
	usw	t9 , - 64(dst)

	subu	size, size, 64
	subu	dst, dst, 64
	subu	src, src, 64
	bgeu	size, 64, 57b

	lw	s0,   0(sp)
	lw	s1,   4(sp)
	lw	s2,   8(sp)
	lw	s3,  12(sp)

/* while ( size > sizeof( ud64_t ) ) {					*/
	
backw_U_32_loop:
	bltu	size, 32, backw_U_16_if

5:
	lw	t0 , -  4(src)
	lw	t1 , -  8(src)
	lw	t2 , - 12(src)
	lw	t3 , - 16(src)
	lw	t4 , - 20(src)
	lw	t5 , - 24(src)
	lw	t6 , - 28(src)
	lw	t7 , - 32(src)
	
	usw	t0 , -  4(dst)
	usw	t1 , -  8(dst)
	usw	t2 , - 12(dst)
	usw	t3 , - 16(dst)
	usw	t4 , - 20(dst)
	usw	t5 , - 24(dst)
	usw	t6 , - 28(dst)
	usw	t7 , - 32(dst)

	subu	size, size, 32
	subu	dst, dst, 32
	subu	src, src, 32
	bgeu	size, 32, 5b

backw_U_16_if:
	bltu	size, 16, backw_U_8_if

	lw	t0 , -  4(src)
	lw	t1 , -  8(src)
	lw	t2 , - 12(src)
	lw	t3 , - 16(src)
	
	usw	t0 , -  4(dst)
	usw	t1 , -  8(dst)
	usw	t2 , - 12(dst)
	usw	t3 , - 16(dst)

	subu	size, size, 16
	subu	dst, dst, 16
	subu	src, src, 16

backw_U_8_if:
	bltu	size, 8, finishit_backw

	lw	t0 , -  4(src)
	lw	t1 , -  8(src)
	
	usw	t0 , -  4(dst)
	usw	t1 , -  8(dst)

	subu	size, size, 8
	subu	dst, dst, 8
	subu	src, src, 8

finishit_backw:

/*  while ( size > sizeof( d4_t ) ) {					*/
/*  ( ( ud4_t * ) dst )[ 0 ] = ( ( ud4_t * ) src )[ 0 ];		*/

	bltu	size, 4, finishit_byte_backw
4:
	ulw	t0, - 4(src)
	usw	t0, - 4(dst)
	
	subu	size, size, 4
	subu	dst, dst, 4
	subu	src, src, 4
	
	bgeu	size, 4, 4b

finishit_byte_backw:

/*  while ( size > 0 ) {						*/
	beq	size, 0, ret
	beq	size, 1, 71f
	beq	size, 2, 72f

	lbu	t0,-1(src)	# copy three bytes
	lbu	t1,-2(src)
	lbu	t2,-3(src)
	sb	t0,-1(dst)
	sb	t1,-2(dst)
	sb	t2,-3(dst)

	b	ret

72:	
	lbu	t0,-1(src)	# copy two bytes
	lbu	t1,-2(src)
	sb	t0,-1(dst)
	sb	t1,-2(dst)

	b	ret

71:
	lbu	t0, - 1(src)
	sb	t0, - 1(dst)
	
	b	ret

END( _bcopy )
