/*
 * ibcopy.s	- Integer only bcopy (as opposed to fp assisted)
 *
 *	Fast memory mover - uses swr/swl instructions for unaligned
 *	copies as well as a big 16 word loop.
 *
 *	This should also work for 64 bit machines.
 *
 *  Lots of ideas taken from Kipp Hickman's original version.
 *
 *  Written by: Gianni Mariani 19-Nov-1993
 */

#include <asm.h>
#include <regdef.h>

#define dst	a0	/* $4 */
#define src	a1	/* $5 */
#define size	a2	/* $6 */

#ifndef t7
#define t4	ta0
#define t5	ta1
#define	t6	ta2
#define t7	ta3
#endif

#ifndef NBPW
#define NBPW		4
#endif
#define STACK_SPACE	(2*NBPW)    # Stack area reserved (must be *2*NBPW)

#if (NBPW != 4)&&(NBPW != 8)
the value of 'NBPW' must be either 4 or 8.
#endif

#if NBPW == 4
#ifndef MIPS2
#define bgezl	bgez
#endif
#endif

#if NBPW == 8
#define lw	ld
#define sw	sd
#define usw	usd
#define ulw	uld
#define lwl	ldl
#define lwr	ldr
#define swl	sdl
#define swr	sdr
#define subu	dsubu
#define	addu	daddu
#endif

LEAF( _bcopy )
	.weakext bcopy, _bcopy

	move	v0, a1		    # move args around to be same as memcpy
	move	src, a0
	move	dst, v0
	bgtz	size, memcpy_start  # Special case bcopy's size is signed
	j	ra
	
XNESTED( memmove )
	move    v0, a0	
	b	memcpy_start

XNESTED( memcpy )
	move	v0, a0

memcpy_start:

	subu    sp, STACK_SPACE	    # make some scratch space
	beq	src, dst, ret	    # get out when src is the same as dst
	
/*  if ( ( dst > src ) && ( dst < ( src + size ) ) ) goto backwards;	*/
	bltu	dst, src, forwards
	addu	v1, src, size
	bltu	dst, v1, backwards

forwards:
/*  if ( size < 4 * sizeof( d4_t ) ) goto finishit;			*/
	bltu	size, 4*NBPW, finishit_forw
	
/*  v7 = ( ( d4_t ) src ) & ( sizeof( d4_t ) - 1 );			*/
/*  v6 = ( ( d4_t ) dst ) & ( sizeof( d4_t ) - 1 );			*/
/*  if ( v6 != v7 ) goto unaligned_forward;				*/

	and	t7, src, NBPW-1
	and	t6, dst, NBPW-1
	bne	t6, t7, unaligned_forward

/* ==================================================================== */
/* ==================================================================== */
/* Copy forward - src and dst are only 4 byte aligned			*/

aligned_forw_A_copy :


	beq	t7, 0, 33f	# align src only if nessasary
	lwl	v1, 0(src)	# read only enough to align
	swl	v1, 0(dst)	# write alignment
	
	addu	size, size, t7
	subu	dst, dst, t7
	subu	src, src, t7
	
	subu	size, size, NBPW
	addu	src, src, NBPW
	addu	dst, dst, NBPW
33:

forw_A_16W_loop:
				    # Don't do the 16W loop unless it will
				    # run at least 4 times
	bltu	size, 16*4*NBPW, forw_A_8W_loop

	# save regs that need saving
	
	mthi	s0		    # save s0 and s1 in the mul regs
	mtlo	s1
	sw	s2,   0*NBPW(sp)    # unfortunatly we cannot use v0
				    # since for memcpy it has the ret val
				    # This is the penalty to bcopy

	subu	size, size, 16*NBPW # make use of the bgez instruction in the
				    # main loop since it removes a slti instr.
				    # theoretically adding 2.7% performance

	/* while ( size >= sizeof( ud64_t ) ) {				*/
	/* do the copy - 4*NBPW bytes at a time				*/
57:
	.set	 noat
	lw	AT ,  0*NBPW(src)
	lw	t1 ,  1*NBPW(src)
	lw	t2 ,  2*NBPW(src)
	lw	t3 ,  3*NBPW(src)
	lw	t4 ,  4*NBPW(src)
	lw	t5 ,  5*NBPW(src)
	lw	t6 ,  6*NBPW(src)
	lw	t7 ,  7*NBPW(src)
	lw	s0 ,  8*NBPW(src)
	lw	s1 ,  9*NBPW(src)
	lw	s2 , 10*NBPW(src)
	lw	t0 , 11*NBPW(src)
	lw	v1 , 12*NBPW(src)
	lw	a3 , 13*NBPW(src)
	lw	t8 , 14*NBPW(src)
	lw	t9 , 15*NBPW(src)
	
	sw	AT ,  0*NBPW(dst)
	.set	 at
	sw	t1 ,  1*NBPW(dst)
	sw	t2 ,  2*NBPW(dst)
	sw	t3 ,  3*NBPW(dst)
	sw	t4 ,  4*NBPW(dst)
	sw	t5 ,  5*NBPW(dst)
	sw	t6 ,  6*NBPW(dst)
	sw	t7 ,  7*NBPW(dst)
	sw	s0 ,  8*NBPW(dst)
	sw	s1 ,  9*NBPW(dst)
	sw	s2 , 10*NBPW(dst)
	sw	t0 , 11*NBPW(dst)
	sw	v1 , 12*NBPW(dst)
	sw	a3 , 13*NBPW(dst)
	sw	t8 , 14*NBPW(dst)
	sw	t9 , 15*NBPW(dst)

	addu	src, src, 16*NBPW
	subu	size, size, 16*NBPW
	addu	dst, dst, 16*NBPW
	bgezl	size, 57b	    # size is 16*NBPW less than the real size

	# Restore registers.
	addu	size, size, 16*NBPW
	mfhi	s0		    # use the mul/div regs as a scratch
	mflo	s1		    # this should be OK here
	lw	s2,   0*NBPW(sp)
	
forw_A_8W_loop:

/*  if ( size > sizeof( 0*NBPW ) ) {					*/
	bltu	size, 8*NBPW, forw_A_4W_if
	
57:
	lw	t0 ,  0*NBPW(src)
	lw	t1 ,  1*NBPW(src)
	lw	t2 ,  2*NBPW(src)
	lw	t3 ,  3*NBPW(src)
	lw	t4 ,  4*NBPW(src)
	lw	t5 ,  5*NBPW(src)
	lw	t6 ,  6*NBPW(src)
	lw	t7 ,  7*NBPW(src)
	
	sw	t0 ,  0*NBPW(dst)
	sw	t1 ,  1*NBPW(dst)
	sw	t2 ,  2*NBPW(dst)
	sw	t3 ,  3*NBPW(dst)
	sw	t4 ,  4*NBPW(dst)
	sw	t5 ,  5*NBPW(dst)
	sw	t6 ,  6*NBPW(dst)
	sw	t7 ,  7*NBPW(dst)
	
	subu	size, size, 8*NBPW
	addu	dst, dst, 8*NBPW
	addu	src, src, 8*NBPW
	
	bgeu	size, 8*NBPW, 57b
	
forw_A_4W_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 4*NBPW, forw_A_2W_if
	
	lw	t0 ,  0*NBPW(src)
	lw	t1 ,  1*NBPW(src)
	lw	t2 ,  2*NBPW(src)
	lw	t3 ,  3*NBPW(src)
	
	sw	t0 ,  0*NBPW(dst)
	sw	t1 ,  1*NBPW(dst)
	sw	t2 ,  2*NBPW(dst)
	sw	t3 ,  3*NBPW(dst)
	
	subu	size, size, 4*NBPW
	addu	dst, dst, 4*NBPW
	addu	src, src, 4*NBPW

forw_A_2W_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 2*NBPW, forw_A_1W_if
	
	lw	t0 ,  0*NBPW(src)
	lw	t1 ,  1*NBPW(src)
	
	sw	t0 ,  0*NBPW(dst)
	sw	t1 ,  1*NBPW(dst)
	
	subu	size, size, 2*NBPW
	addu	dst, dst, 2*NBPW
	addu	src, src, 2*NBPW

forw_A_1W_if:

/*  aligned_forw_finish :						*/
/*  while ( size >= sizeof( d4_t ) ) {					*/
	bltu	size, 1*NBPW, aligned_forw_partial
63:

	lw	t0, 0(src)
	sw	t0, 0(dst)

	subu	size, size, 1*NBPW
	addu	src, src, 1*NBPW
	addu	dst, dst, 1*NBPW

aligned_forw_partial:

	beq     size, 0, ret

	addu	src, src, size
	addu	dst, dst, size

	lwr	t0, -1(src)
	swr	t0, -1(dst)

ret:
	addu	sp, STACK_SPACE
	j	ra

/* ==================================================================== */
/* ==================================================================== */
/* Copy forward - src and dst are not aligned				*/

unaligned_forward:

#if NBPW == 4

	# In this case t7 contains NBPW-size to copy - do a binary like
	# search. This makes it two comparisons for every size.  It is also
	# a little less code.
	
	beq	t7, zero, 33f	# align src only if nessasary
	
	lbu	t0, 0(src)	# at least one byte there
	beq     t7, 3, 21f
	
	lbu	t1, 1(src)	# at least two bytes there etc	
	beq     t7, 2, 22f
	
	lbu	t2, 2(src)
	sb	t0, 0(dst)
	sb	t1, 1(dst)
	sb	t2, 2(dst)
	subu	size, size, 3
	addu	dst, dst, 3
	addu	src, src, 3
	b	33f
22:

	sb	t0, 0(dst)
	sb	t1, 1(dst)
	subu	size, size, 2
	addu	dst, dst, 2
	addu	src, src, 2
	b	33f

21:
	sb	t0, 0(dst)
	
	subu	size, size, 1
	addu	dst, dst, 1
	addu	src, src, 1
33:
#endif

/* Select the rest of the cases for unaligned copies when NBPW == 8	*/
#if NBPW == 8
	# In this case t7 contains NBPW-size to copy - do a binary like
	# search. This makes it three comparisons for every size.  It is also
	# a little less code.
	
	beq     t7, zero, 33f	# Test this first since its dependably fast
	
	lbu	t0, 0(src)	# at least one byte there	
	bleu	t7, NBPW-4, 44f
	
	bne	t7, NBPW-1, 42f	# align src only if nessasary

	sb      t0, 0(dst)
	subu	size, size, 1
	addu	dst, dst, 1
	addu	src, src, 1
	b	33f

42:
	lbu     t1, 1(src)
	sb	t0, 0(dst)
	sb	t1, 1(dst)
	bne     t7, NBPW-3, 52f

	lbu     t2, 2(src)	# Must be 3
	sb	t2, 2(dst)
	subu	size, size, 3
	addu	dst, dst, 3
	addu	src, src, 3
	b	33f

52:
	subu	size, size, 2
	addu	dst, dst, 2
	addu	src, src, 2
	b	33f

44:
	lbu     t1, 1(src)	# at least 4 bytes there	
	lbu     t2, 2(src)
	lbu     t3, 3(src)

	sb	t0, 0(dst)
	sb	t1, 1(dst)
	sb	t2, 2(dst)
	sb	t3, 3(dst)
	
	bleu	t7, NBPW-6, 46f
	beq     t7, NBPW-4, 54f

	# must be 5 bytes
	lbu	t4, 4(src)	# at least one byte there
	sb	t4, 4(dst)
	subu	size, size, 5
	addu	dst, dst, 5
	addu	src, src, 5
	b	33f
54:
	subu	size, size, 4
	addu	dst, dst, 4
	addu	src, src, 4
	b	33f

46:
	lbu	t4, 4(src)	# at least 6 bytes
	lbu	t5, 5(src)
	sb	t4, 4(dst)
	sb	t5, 5(dst)
	beq	t7, NBPW-6, 56f
	
	lbu	t6, 6(src)
	sb	t6, 6(dst)
	subu	size, size, 7
	addu	dst, dst, 7
	addu	src, src, 7
	b	33f

56:
	subu	size, size, 6
	addu	dst, dst, 6
	addu	src, src, 6

33:

#endif


forw_U_16W_loop:
				    # Don't do the 16W loop unless it will
				    # run at least 4 times
	bltu	size, 16*4*NBPW, forw_U_8W_loop

	# save regs that need saving
	
	mthi	s0		    # save s0 and s1 in the mul regs
	mtlo	s1
	sw	s2,   0*NBPW(sp)    # unfortunatly we cannot use v0
				    # since for memcpy it has the ret val
				    # This is the penalty to bcopy

	subu	size, size, 16*NBPW # make use of the bgez instruction in the
				    # main loop since it removes a slti instr.
				    # theoretically adding 2.7% performance

	/* while ( size >= sizeof( ud64_t ) ) {				*/
	/* do the copy - 16*NBPW bytes at a time				*/
57:
	.set	 noat
	lw	AT ,  0*NBPW(src)
	lw	t1 ,  1*NBPW(src)
	lw	t2 ,  2*NBPW(src)
	lw	t3 ,  3*NBPW(src)
	lw	t4 ,  4*NBPW(src)
	lw	t5 ,  5*NBPW(src)
	lw	t6 ,  6*NBPW(src)
	lw	t7 ,  7*NBPW(src)
	lw	s0 ,  8*NBPW(src)
	lw	s1 ,  9*NBPW(src)
	lw	s2 , 10*NBPW(src)
	lw	t0 , 11*NBPW(src)
	lw	v1 , 12*NBPW(src)
	lw	a3 , 13*NBPW(src)
	lw	t8 , 14*NBPW(src)
	lw	t9 , 15*NBPW(src)
	
	usw	AT ,  0*NBPW(dst)
	.set	 at
	usw	t1 ,  1*NBPW(dst)
	usw	t2 ,  2*NBPW(dst)
	usw	t3 ,  3*NBPW(dst)
	usw	t4 ,  4*NBPW(dst)
	usw	t5 ,  5*NBPW(dst)
	usw	t6 ,  6*NBPW(dst)
	usw	t7 ,  7*NBPW(dst)
	usw	s0 ,  8*NBPW(dst)
	usw	s1 ,  9*NBPW(dst)
	usw	s2 , 10*NBPW(dst)
	usw	t0 , 11*NBPW(dst)
	usw	v1 , 12*NBPW(dst)
	usw	a3 , 13*NBPW(dst)
	usw	t8 , 14*NBPW(dst)
	usw	t9 , 15*NBPW(dst)

	addu	src, src, 16*NBPW
	subu	size, size, 16*NBPW
	addu	dst, dst, 16*NBPW
	bgezl	size, 57b	    # size is 16*NBPW less than the real size

	# Restore registers.
	addu	size, size, 16*NBPW
	mfhi	s0		    # use the mul/div regs as a scratch
	mflo	s1		    # this should be OK here
	lw	s2,   0*NBPW(sp)
	
forw_U_8W_loop:
	bltu	size, 8*NBPW, forw_U_4W_if

5:
	lw	t0 , 0*NBPW(src)
	lw	t1 , 1*NBPW(src)
	lw	t2 , 2*NBPW(src)
	lw	t3 , 3*NBPW(src)
	lw	t4 , 4*NBPW(src)
	lw	t5 , 5*NBPW(src)
	lw	t6 , 6*NBPW(src)
	lw	t7 , 7*NBPW(src)
	
	usw	t0 , 0*NBPW(dst)
	usw	t1 , 1*NBPW(dst)
	usw	t2 , 2*NBPW(dst)
	usw	t3 , 3*NBPW(dst)
	usw	t4 , 4*NBPW(dst)
	usw	t5 , 5*NBPW(dst)
	usw	t6 , 6*NBPW(dst)
	usw	t7 , 7*NBPW(dst)

	subu	size, size, 8*NBPW
	addu	dst, dst, 8*NBPW
	addu	src, src, 8*NBPW
	bgeu	size, 8*NBPW, 5b

forw_U_4W_if:
	bltu	size, 4*NBPW, forw_U_2W_if

	lw	t0 , 0*NBPW(src)
	lw	t1 , 1*NBPW(src)
	lw	t2 , 2*NBPW(src)
	lw	t3 , 3*NBPW(src)
	
	usw	t0 , 0*NBPW(dst)
	usw	t1 , 1*NBPW(dst)
	usw	t2 , 2*NBPW(dst)
	usw	t3 , 3*NBPW(dst)

	subu	size, size, 4*NBPW
	addu	dst, dst, 4*NBPW
	addu	src, src, 4*NBPW

forw_U_2W_if:
	bltu	size, 2*NBPW, finishit_forw

	lw	t0 ,  0*NBPW(src)
	lw	t1 ,  1*NBPW(src)
	
	usw	t0 ,  0*NBPW(dst)
	usw	t1 ,  1*NBPW(dst)

	subu	size, size, 2*NBPW
	addu	dst, dst, 2*NBPW
	addu	src, src, 2*NBPW

finishit_forw:

/*  while ( size >= sizeof( word ) ) {					*/

	bltu	size, 1*NBPW, finishit_forw_byte
4:
	ulw	t0, 0(src)
	usw	t0, 0(dst)

	subu	size, size, 1*NBPW
	addu	dst, dst, 1*NBPW
	addu	src, src, 1*NBPW
	
	bgeu	size, 1*NBPW, 4b

finishit_forw_byte:

#if NBPW == 4

	# In this case t7 contains NBPW-size to copy - do a binary like
	# search. This makes it two comparisons for every size.  It is also
	# a little less code.
	
	beq	size, zero, ret	# if zero - get out
	
	lbu	t0, 0(src)	# at least one byte there
	sb	t0, 0(dst)
	beq     size, 1, ret
	
	lbu	t1, 1(src)	# at least two bytes there etc	
	sb	t1, 1(dst)
	
	beq     size, 2, ret
	
	lbu	t2, 2(src)
	sb	t2, 2(dst)
	b	ret
#endif

#if NBPW == 8
	beq     size, 0, ret	#
	lbu	t0, 0(src)	# at least one byte there
	bgeu	size, 4, 44f
	sb      t0, 0(dst)
	beq	size, 1, ret	# align src only if nessasary
	
	lbu     t1, 1(src)
	sb	t1, 1(dst)
	bne     size, 3, ret

	lbu     t2, 2(src)	# Must be 3
	sb	t2, 2(dst)
	b	ret

44:
	lbu     t1, 1(src)
	lbu     t2, 2(src)
	lbu     t3, 3(src)

	sb      t0, 0(dst)
	sb	t1, 1(dst)
	sb	t2, 2(dst)
	sb	t3, 3(dst)
	
	bgeu	size, 6, 46f
	beq     size, 4, ret

	# must be 5 bytes
	lbu	t4, 4(src)	# at least one byte there
	sb	t4, 4(dst)
	b	ret

46:
	lbu	t4, 4(src)	# at least 6 bytes
	lbu	t5, 5(src)
	sb	t4, 4(dst)
	sb	t5, 5(dst)
	beq	size, 6, ret
	
	lbu	t6, 6(src)
	sb	t6, 6(dst)
	b	ret

#endif

/* ==================================================================== */
/* ==================================================================== */
/* Backwards copies							*/

backwards:

/* Point to the end ( and beyond by one byte )				*/
	addu	dst, dst, size
	addu	src, src, size

/*  if ( size < 4 * sizeof( d4_t ) ) goto finishit;			*/
	bltu	size, 4*NBPW, finishit_backw
	
/*  v7 = ( ( d4_t ) src ) & ( sizeof( d4_t ) - 1 );			*/
/*  v6 = ( ( d4_t ) dst ) & ( sizeof( d4_t ) - 1 );			*/
/*  if ( v6 != v7 ) goto unaligned_backward;				*/
	and	t6, dst, NBPW-1
	and	t7, src, NBPW-1

	bne	t6, t7, unaligned_backward

/* ==================================================================== */
/* ==================================================================== */
/* Copy backward - src and dst are only 4 byte aligned			*/

aligned_backw_A_copy :


	beq	t7, 0, 33f	# align src only if nessasary
	lwr	v1, -1(src)	# read only enough to align
	swr	v1, -1(dst)	# write alignment
	
	subu	size, size, t7
	subu	dst, dst, t7
	subu	src, src, t7
	
33:
backw_A_16W_loop:

				    # Don't do the 16W loop unless it will
				    # run at least 4 times
	bltu	size, 16*4*NBPW, backw_A_8W_loop

	# save regs that need saving
	
	mthi	s0		    # save s0 and s1 in the mul regs
	mtlo	s1

	subu	size, size, 16*NBPW # make use of the bgez instruction in the
				    # main loop since it removes a slti instr.
				    # theoretically adding 2.7% performance

	/* while ( size >= sizeof( ud64_t ) ) {				*/
	/* do the copy - 4*NBPW bytes at a time				*/
57:
	.set	 noat
	lw	AT , - ( 1 +  0 )*NBPW(src)
	lw	t1 , - ( 1 +  1 )*NBPW(src)
	lw	t2 , - ( 1 +  2 )*NBPW(src)
	lw	t3 , - ( 1 +  3 )*NBPW(src)
	lw	t4 , - ( 1 +  4 )*NBPW(src)
	lw	t5 , - ( 1 +  5 )*NBPW(src)
	lw	t6 , - ( 1 +  6 )*NBPW(src)
	lw	t7 , - ( 1 +  7 )*NBPW(src)
	lw	s0 , - ( 1 +  8 )*NBPW(src)
	lw	s1 , - ( 1 +  9 )*NBPW(src)
	lw	v0 , - ( 1 + 10 )*NBPW(src)
	lw	t0 , - ( 1 + 11 )*NBPW(src)
	lw	v1 , - ( 1 + 12 )*NBPW(src)
	lw	a3 , - ( 1 + 13 )*NBPW(src)
	lw	t8 , - ( 1 + 14 )*NBPW(src)
	lw	t9 , - ( 1 + 15 )*NBPW(src)
	
	sw	AT , - ( 1 +  0 )*NBPW(dst)
	.set	 at
	sw	t1 , - ( 1 +  1 )*NBPW(dst)
	sw	t2 , - ( 1 +  2 )*NBPW(dst)
	sw	t3 , - ( 1 +  3 )*NBPW(dst)
	sw	t4 , - ( 1 +  4 )*NBPW(dst)
	sw	t5 , - ( 1 +  5 )*NBPW(dst)
	sw	t6 , - ( 1 +  6 )*NBPW(dst)
	sw	t7 , - ( 1 +  7 )*NBPW(dst)
	sw	s0 , - ( 1 +  8 )*NBPW(dst)
	sw	s1 , - ( 1 +  9 )*NBPW(dst)
	sw	v0 , - ( 1 + 10 )*NBPW(dst)
	sw	t0 , - ( 1 + 11 )*NBPW(dst)
	sw	v1 , - ( 1 + 12 )*NBPW(dst)
	sw	a3 , - ( 1 + 13 )*NBPW(dst)
	sw	t8 , - ( 1 + 14 )*NBPW(dst)
	sw	t9 , - ( 1 + 15 )*NBPW(dst)

	subu	src, src, 16*NBPW
	subu	size, size, 16*NBPW
	subu	dst, dst, 16*NBPW
	bgezl	size, 57b	    # size is 16*NBPW less than the real size

	# Restore registers.
	addu	size, size, 16*NBPW
	mfhi	s0		    # use the mul/div regs as a scratch
	mflo	s1		    # this should be OK here
	subu	v0, dst, size	    # v0 is allways this in back copies
	
backw_A_8W_loop:

/*  if ( size > sizeof( 0*NBPW ) ) {					*/
	bltu	size, 8*NBPW, backw_A_4W_if
	
57:
	lw	t0 , - ( 1 +  0 )*NBPW(src)
	lw	t1 , - ( 1 +  1 )*NBPW(src)
	lw	t2 , - ( 1 +  2 )*NBPW(src)
	lw	t3 , - ( 1 +  3 )*NBPW(src)
	lw	t4 , - ( 1 +  4 )*NBPW(src)
	lw	t5 , - ( 1 +  5 )*NBPW(src)
	lw	t6 , - ( 1 +  6 )*NBPW(src)
	lw	t7 , - ( 1 +  7 )*NBPW(src)
	
	sw	t0 , - ( 1 +  0 )*NBPW(dst)
	sw	t1 , - ( 1 +  1 )*NBPW(dst)
	sw	t2 , - ( 1 +  2 )*NBPW(dst)
	sw	t3 , - ( 1 +  3 )*NBPW(dst)
	sw	t4 , - ( 1 +  4 )*NBPW(dst)
	sw	t5 , - ( 1 +  5 )*NBPW(dst)
	sw	t6 , - ( 1 +  6 )*NBPW(dst)
	sw	t7 , - ( 1 +  7 )*NBPW(dst)
	
	subu	size, size, 8*NBPW
	subu	dst, dst, 8*NBPW
	subu	src, src, 8*NBPW
	
	bgeu	size, 8*NBPW, 57b
	
backw_A_4W_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 4*NBPW, backw_A_2W_if
	
	lw	t0 , - ( 1 +  0 )*NBPW(src)
	lw	t1 , - ( 1 +  1 )*NBPW(src)
	lw	t2 , - ( 1 +  2 )*NBPW(src)
	lw	t3 , - ( 1 +  3 )*NBPW(src)
	
	sw	t0 , - ( 1 +  0 )*NBPW(dst)
	sw	t1 , - ( 1 +  1 )*NBPW(dst)
	sw	t2 , - ( 1 +  2 )*NBPW(dst)
	sw	t3 , - ( 1 +  3 )*NBPW(dst)
	
	subu	size, size, 4*NBPW
	subu	dst, dst, 4*NBPW
	subu	src, src, 4*NBPW

backw_A_2W_if:

/*  if ( size > sizeof( ud16_t ) ) {					*/
	bltu	size, 2*NBPW, backw_A_1W_if
	
	lw	t0 , - ( 1 +  0 )*NBPW(src)
	lw	t1 , - ( 1 +  1 )*NBPW(src)
	
	sw	t0 , - ( 1 +  0 )*NBPW(dst)
	sw	t1 , - ( 1 +  1 )*NBPW(dst)
	
	subu	size, size, 2*NBPW
	subu	dst, dst, 2*NBPW
	subu	src, src, 2*NBPW

backw_A_1W_if:

/*  aligned_backw_finish :						*/
/*  if ( size >= sizeof( d4_t ) ) {					*/
	bltu	size, 1*NBPW, aligned_backw_partial
63:

	lw	t0, - ( 1 + 0 )*NBPW(src)
	sw	t0, - ( 1 + 0 )*NBPW(dst)

	subu	size, size, 1*NBPW
	subu	src, src, 1*NBPW
	subu	dst, dst, 1*NBPW

aligned_backw_partial:

	beq     size, 0, ret

	subu	src, src, size
	subu	dst, dst, size

	lwl	t0, 0(src)
	swl	t0, 0(dst)

	b	ret

/* ==================================================================== */
/* ==================================================================== */
/* Copy backward - src and dst are not aligned				*/

unaligned_backward:

#if NBPW == 4

	# In this case t7 contains NBPW-size to copy - do a binary like
	# search. This makes it two comparisons for every size.  It is also
	# a little less code.
	
	beq	t7, zero, 33f	# align src only if nessasary
	
	lbu	t0, -(1+0)(src)	# at least one byte there
	beq     t7, 1, 21f
	
	lbu	t1, -(1+1)(src)	# at least two bytes there etc	
	beq     t7, 2, 22f
	
	lbu	t2, -(1+2)(src)
	sb	t0, -(1+0)(dst)
	sb	t1, -(1+1)(dst)
	sb	t2, -(1+2)(dst)
	subu	size, size, 3
	subu	dst, dst, 3
	subu	src, src, 3
	b	33f
22:

	sb	t0, -(1+0)(dst)
	sb	t1, -(1+1)(dst)
	subu	size, size, 2
	subu	dst, dst, 2
	subu	src, src, 2
	b	33f

21:
	sb	t0, -(1+0)(dst)
	
	subu	size, size, 1
	subu	dst, dst, 1
	subu	src, src, 1
33:
#endif

/* Select the rest of the cases for unaligned copies when NBPW == 8	*/
#if NBPW == 8
	# In this case t7 contains the size to copy - do a binary like
	# search. This makes it three comparisons for every size.  It is also
	# a little less code.
	
	beq     t7, zero, 33f	# Test this first since its dependably fast
	
	lbu	t0, -(1+0)(src)	# at least one byte there	
	bgeu	t7, 4, 44f
	
	bne	t7, 1, 42f	# align src only if nessasary

	sb      t0, -(1+0)(dst)
	subu	size, size, 1
	subu	dst, dst, 1
	subu	src, src, 1
	b	33f

42:
	lbu     t1, -(1+1)(src)
	sb	t0, -(1+0)(dst)
	sb	t1, -(1+1)(dst)
	bne     t7, 3, 52f

	lbu     t2, -(1+2)(src)	# Must be 3
	sb	t2, -(1+2)(dst)
	subu	size, size, 3
	subu	dst, dst, 3
	subu	src, src, 3
	b	33f

52:
	subu	size, size, 2
	subu	dst, dst, 2
	subu	src, src, 2
	b	33f

44:
	lbu     t1, -(1+1)(src)	# at least 4 bytes there	
	lbu     t2, -(1+2)(src)
	lbu     t3, -(1+3)(src)

	sb	t0, -(1+0)(dst)
	sb	t1, -(1+1)(dst)
	sb	t2, -(1+2)(dst)
	sb	t3, -(1+3)(dst)
	
	bgeu	t7, 6, 46f
	beq     t7, 4, 54f

	# must be 5 bytes
	lbu	t4, -(1+4)(src)	# at least one byte there
	sb	t4, -(1+4)(dst)
	subu	size, size, 5
	subu	dst, dst, 5
	subu	src, src, 5
	b	33f
54:
	subu	size, size, 4
	subu	dst, dst, 4
	subu	src, src, 4
	b	33f

46:
	lbu	t4, -(1+4)(src)	# at least 6 bytes
	lbu	t5, -(1+5)(src)
	sb	t4, -(1+4)(dst)
	sb	t5, -(1+5)(dst)
	beq	t7, 6, 56f
	
	lbu	t6, -(1+6)(src)
	sb	t6, -(1+6)(dst)
	subu	size, size, 7
	subu	dst, dst, 7
	subu	src, src, 7
	b	33f

56:
	subu	size, size, 6
	subu	dst, dst, 6
	subu	src, src, 6

33:

#endif


backw_U_16W_loop:
				    # Don't do the 16W loop unless it will
				    # run at least 4 times
	bltu	size, 16*4*NBPW, backw_U_8W_loop

	# save regs that need saving
	
	mthi	s0		    # save s0 and s1 in the mul regs
	mtlo	s1
	subu	size, size, 16*NBPW # make use of the bgez instruction in the
				    # main loop since it removes a slti instr.
				    # theoretically adding 2.7% performance

	/* while ( size >= sizeof( ud64_t ) ) {				*/
	/* do the copy - 16*NBPW bytes at a time				*/
57:
	.set	 noat
	lw	AT , -( 1 +  0 )*NBPW(src)
	lw	t1 , -( 1 +  1 )*NBPW(src)
	lw	t2 , -( 1 +  2 )*NBPW(src)
	lw	t3 , -( 1 +  3 )*NBPW(src)
	lw	t4 , -( 1 +  4 )*NBPW(src)
	lw	t5 , -( 1 +  5 )*NBPW(src)
	lw	t6 , -( 1 +  6 )*NBPW(src)
	lw	t7 , -( 1 +  7 )*NBPW(src)
	lw	s0 , -( 1 +  8 )*NBPW(src)
	lw	s1 , -( 1 +  9 )*NBPW(src)
	lw	v0 , -( 1 + 10 )*NBPW(src)
	lw	t0 , -( 1 + 11 )*NBPW(src)
	lw	v1 , -( 1 + 12 )*NBPW(src)
	lw	a3 , -( 1 + 13 )*NBPW(src)
	lw	t8 , -( 1 + 14 )*NBPW(src)
	lw	t9 , -( 1 + 15 )*NBPW(src)
	
	usw	AT , -( 1 +  0 )*NBPW(dst)
	.set	 at
	usw	t1 , -( 1 +  1 )*NBPW(dst)
	usw	t2 , -( 1 +  2 )*NBPW(dst)
	usw	t3 , -( 1 +  3 )*NBPW(dst)
	usw	t4 , -( 1 +  4 )*NBPW(dst)
	usw	t5 , -( 1 +  5 )*NBPW(dst)
	usw	t6 , -( 1 +  6 )*NBPW(dst)
	usw	t7 , -( 1 +  7 )*NBPW(dst)
	usw	s0 , -( 1 +  8 )*NBPW(dst)
	usw	s1 , -( 1 +  9 )*NBPW(dst)
	usw	v0 , -( 1 + 10 )*NBPW(dst)
	usw	t0 , -( 1 + 11 )*NBPW(dst)
	usw	v1 , -( 1 + 12 )*NBPW(dst)
	usw	a3 , -( 1 + 13 )*NBPW(dst)
	usw	t8 , -( 1 + 14 )*NBPW(dst)
	usw	t9 , -( 1 + 15 )*NBPW(dst)

	subu	src, src, 16*NBPW
	subu	size, size, 16*NBPW
	subu	dst, dst, 16*NBPW
	bgezl	size, 57b	    # size is 16*NBPW less than the real size

	# Restore registers.
	addu	size, size, 16*NBPW
	mfhi	s0		    # use the mul/div regs as a scratch
	mflo	s1		    # this should be OK here
	subu	v0, dst, size	    # v0 is allways this in back copies

backw_U_8W_loop:
	bltu	size, 8*NBPW, backw_U_4W_if

5:
	lw	t0 , -( 1 + 0 )*NBPW(src)
	lw	t1 , -( 1 + 1 )*NBPW(src)
	lw	t2 , -( 1 + 2 )*NBPW(src)
	lw	t3 , -( 1 + 3 )*NBPW(src)
	lw	t4 , -( 1 + 4 )*NBPW(src)
	lw	t5 , -( 1 + 5 )*NBPW(src)
	lw	t6 , -( 1 + 6 )*NBPW(src)
	lw	t7 , -( 1 + 7 )*NBPW(src)
	
	usw	t0 , -( 1 + 0 )*NBPW(dst)
	usw	t1 , -( 1 + 1 )*NBPW(dst)
	usw	t2 , -( 1 + 2 )*NBPW(dst)
	usw	t3 , -( 1 + 3 )*NBPW(dst)
	usw	t4 , -( 1 + 4 )*NBPW(dst)
	usw	t5 , -( 1 + 5 )*NBPW(dst)
	usw	t6 , -( 1 + 6 )*NBPW(dst)
	usw	t7 , -( 1 + 7 )*NBPW(dst)

	subu	size, size, 8*NBPW
	subu	dst, dst, 8*NBPW
	subu	src, src, 8*NBPW
	bgeu	size, 8*NBPW, 5b

backw_U_4W_if:
	bltu	size, 4*NBPW, backw_U_2W_if

	lw	t0 , -( 1 + 0 )*NBPW(src)
	lw	t1 , -( 1 + 1 )*NBPW(src)
	lw	t2 , -( 1 + 2 )*NBPW(src)
	lw	t3 , -( 1 + 3 )*NBPW(src)
	
	usw	t0 , -( 1 + 0 )*NBPW(dst)
	usw	t1 , -( 1 + 1 )*NBPW(dst)
	usw	t2 , -( 1 + 2 )*NBPW(dst)
	usw	t3 , -( 1 + 3 )*NBPW(dst)

	subu	size, size, 4*NBPW
	subu	dst, dst, 4*NBPW
	subu	src, src, 4*NBPW

backw_U_2W_if:
	bltu	size, 2*NBPW, finishit_backw

	lw	t0 , -( 1 +  0 )*NBPW(src)
	lw	t1 , -( 1 +  1 )*NBPW(src)
	
	usw	t0 , -( 1 +  0 )*NBPW(dst)
	usw	t1 , -( 1 +  1 )*NBPW(dst)

	subu	size, size, 2*NBPW
	subu	dst, dst, 2*NBPW
	subu	src, src, 2*NBPW

finishit_backw:

/*  while ( size >= sizeof( word ) ) {					*/

	bltu	size, 1*NBPW, finishit_backw_byte
4:
	ulw	t0, -( 1 + 0 )*NBPW(src)
	usw	t0, -( 1 + 0 )*NBPW(dst)

	subu	size, size, 1*NBPW
	subu	dst, dst, 1*NBPW
	subu	src, src, 1*NBPW
	
	bgeu	size, 1*NBPW, 4b

finishit_backw_byte:

#if NBPW == 4

	# This makes it two comparisons for every size.  It is also
	# a little less code.
	
	beq	size, zero, ret	# if zero - get out
	
	lbu	t0, -(1+0)(src)	# at least one byte there
	sb	t0, -(1+0)(dst)
	beq     size, 1, ret
	
	lbu	t1, -(1+1)(src)	# at least two bytes there etc	
	sb	t1, -(1+1)(dst)
	
	beq     size, 2, ret
	
	lbu	t2, -(1+2)(src)
	sb	t2, -(1+2)(dst)
	b	ret
#endif

#if NBPW == 8
	beq     size, 0, ret	#
	lbu	t0, -(1+0)(src)	# at least one byte there
	bgeu	size, 4, 44f
	sb      t0, -(1+0)(dst)
	beq	size, 1, ret	# align src only if nessasary
	
	lbu     t1, -(1+1)(src)
	sb	t1, -(1+1)(dst)
	bne     size, 3, ret

	lbu     t2, -(1+2)(src)	# Must be 3
	sb	t2, -(1+2)(dst)
	b	ret

44:
	lbu     t1, -(1+1)(src)
	lbu     t2, -(1+2)(src)
	lbu     t3, -(1+3)(src)

	sb      t0, -(1+0)(dst)
	sb	t1, -(1+1)(dst)
	sb	t2, -(1+2)(dst)
	sb	t3, -(1+3)(dst)
	
	bgeu	size, 6, 46f
	beq     size, 4, ret

	# must be 5 bytes
	lbu	t4, -(1+4)(src)	# at least one byte there
	sb	t4, -(1+4)(dst)
	b	ret

46:
	lbu	t4, -(1+4)(src)	# at least 6 bytes
	lbu	t5, -(1+5)(src)
	sb	t4, -(1+4)(dst)
	sb	t5, -(1+5)(dst)
	beq	size, 6, ret
	
	lbu	t6, -(1+6)(src)
	sb	t6, -(1+6)(dst)
	b	ret

#endif


END( _bcopy )
